#version 400

layout (std140) uniform Lights
{
	vec3 light_position;		// In camera space
	vec3 light_amb;
	vec3 light_dif;
	vec3 light_spe;
};

layout (std140) uniform Material
{
	vec3 mat_amb;
	vec3 mat_dif;
	vec3 mat_spe;
	float mat_shi;
};

uniform sampler2D bumpTex;

uniform float height_scale;

in Data
{
	vec3 Position;
	vec3 Normal;
	vec3 Tangent;
	vec3 Bitangent;
	vec2 TexCoord;
} DataIn;

out vec4 outputF;

// Computes factors for Blinn-Phong lighting model. Similar to 'lit' in HLSL
vec4 lit(in float n_dot_l, in float n_dot_h, in float exponent)
{
	if (n_dot_l < 0.0)
		return vec4(1.0, 0.0, 0.0, 1.0);
	else
		if (n_dot_h < 0.0)
			return vec4(1.0, n_dot_l, 0.0, 1.0);
		else
			return vec4(1.0, n_dot_l, pow(n_dot_h, exponent), 1.0);
}

//----------------------------------------------------------------
// TODO:
// 7) Comment single black color, uncomment Phong shading
// 10) Comment fixed normal, uncomment bump mapping
//----------------------------------------------------------------

void main()
{
	// Black only
	// TODO 7) Comment single black color, uncomment Phong shading
	//outputF = vec4(0.0);

	// TODO 10) Comment fixed normal, uncomment bump mapping
	// Compute the derivative of the height texture. Try higher LOD or lower approximation step
	// to reduce the artefacts.
	float step_lod = 4.0;
	float height_h = 1.0 / float(textureSize(bumpTex, 0)) * step_lod;
	float sample_lod = 0.0;
	float height = textureLod(bumpTex, DataIn.TexCoord, sample_lod).x;
	float height_dx = (textureLod(bumpTex, vec2(DataIn.TexCoord.x + height_h, DataIn.TexCoord.y), sample_lod).x - height) / height_h;
	float height_dy = (textureLod(bumpTex, vec2(DataIn.TexCoord.x, DataIn.TexCoord.y + height_h), sample_lod).x - height) / height_h;

	vec3 t = DataIn.Tangent + DataIn.Normal * height_dx * height_scale;
	vec3 b = DataIn.Bitangent + DataIn.Normal * height_dy * height_scale;
	vec3 n = normalize(cross(t, b));
	
	// TODO 7) Comment single black color, uncomment Phong shading
	// TODO 10) Comment fixed normal, uncomment bump mapping
	//vec3 n = normalize(Normal);

	vec3 light_vec = normalize(light_position - DataIn.Position);
	vec3 view_vec = normalize(-DataIn.Position);
	vec3 half_vec = normalize(light_vec + view_vec);

	vec4 factors = lit(dot(n, light_vec), dot(n, half_vec), mat_shi);
	vec3 color =
		light_amb * mat_amb * factors.x +
		light_dif * mat_dif * factors.y +
		light_spe * mat_spe * factors.z;

	outputF = vec4(color, 1.0);
}