#version 400

layout(vertices = 16) out;

layout (std140) uniform Matrices
{
	mat4 m_model;
	mat4 m_view;
	mat4 m_projection;
	mat4 m_pvm;
};

uniform float win_width;
uniform float win_height;

// Approximative size of one triangle in pixels
const float triangle_size = 20.0f;

in Data
{
	vec4 position;
} DataIn[];

out Data
{
	vec4 position;
} DataOut[];

// Only to pass the data between invocations, not used in the evaluation shader
out vec4 screens[];

//----------------------------------------------------------------
// TODO:
// 6) Pass position from VS to TES
// 6) Set all tessellation levels to 30.0f (set them only from gl_InvocationID == 0)
// 11) Comment setting all tessellation levels to 30.0f
// 11) Compute screen position of all vertices and use LOD tessellation
// 12) Try simple view frustum culling
//----------------------------------------------------------------

void main()
{
	// TODO 6) Pass position from VS to TES
	DataOut[gl_InvocationID].position = DataIn[gl_InvocationID].position;

	// TODO 6) Set all tessellation levels to 30.0f (set them only from gl_InvocationID == 0)
	// TODO 11) Comment setting all tessellation levels to 30.0f
	//if (gl_InvocationID == 0)
	//{
	//	gl_TessLevelInner[0] = 30.0f;
	//	gl_TessLevelInner[1] = 30.0f;
	//	gl_TessLevelOuter[0] = 30.0f;
	//	gl_TessLevelOuter[1] = 30.0f;
	//	gl_TessLevelOuter[2] = 30.0f;
	//	gl_TessLevelOuter[3] = 30.0f;
	//}
	//return;

	// TODO 11) Compute screen position of all vertices and use LOD tessellation
	vec4 screen_position = m_projection * DataIn[gl_InvocationID].position;
	screens[gl_InvocationID] = screen_position / screen_position.w;
	barrier();

	if (gl_InvocationID == 0)
	{
		bool culled = false;
		// TODO 12) Try simple view frustum culling
		if (false)		// Do primitive culling
		{
			// Compute the bounding box of all primitives
			vec4 bb_min = vec4(1000.0);
			vec4 bb_max = vec4(-1000.0);
			for (int i = 0; i < 16; i++)
			{
				bb_min = min(bb_min, screens[i]);
				bb_max = max(bb_max, screens[i]);
			}

			// Proper culling would cull xyz against volume (-1.0, 1.0).
			// However, we would not see any result, so cull xy againts rectangle (-0.5, 0.5).
			vec2 culling_rect = vec2(0.5);

			if (any(lessThan(bb_max.xy, -culling_rect)) || any(greaterThan(bb_min.xy, culling_rect)))
			{
				// The whole patch is out of the frustum
				culled = true;
			}
		}

		if (culled)
		{
			// No tessellation
			gl_TessLevelInner[0] = 0.0;
			gl_TessLevelInner[1] = 0.0;
			gl_TessLevelOuter[0] = 0.0;
			gl_TessLevelOuter[1] = 0.0;
			gl_TessLevelOuter[2] = 0.0;
			gl_TessLevelOuter[3] = 0.0;
		}
		else
		{
			if (false)
			{
				// Constant tessellation
				gl_TessLevelInner[0] = 30.0f;
				gl_TessLevelInner[1] = 30.0f;
				gl_TessLevelOuter[0] = 30.0f;
				gl_TessLevelOuter[1] = 30.0f;
				gl_TessLevelOuter[2] = 30.0f;
				gl_TessLevelOuter[3] = 30.0f;
			}
			else
			{
				// Compute approximative length of each side of the bezier patch in pixels ...
				vec2 win_size = vec2(win_width, win_height);
				vec2 p0_0 = (screens[8] - screens[12]).xy * win_size;
				vec2 p0_1 = (screens[4] - screens[8]).xy * win_size;
				vec2 p0_2 = (screens[0] - screens[4]).xy * win_size;
				vec2 p1_0 = (screens[1] - screens[0]).xy * win_size;
				vec2 p1_1 = (screens[2] - screens[1]).xy * win_size;
				vec2 p1_2 = (screens[3] - screens[2]).xy * win_size;
				vec2 p2_0 = (screens[7] - screens[3]).xy * win_size;
				vec2 p2_1 = (screens[11] - screens[7]).xy * win_size;
				vec2 p2_2 = (screens[15] - screens[11]).xy * win_size;
				vec2 p3_0 = (screens[14] - screens[15]).xy * win_size;
				vec2 p3_1 = (screens[13] - screens[14]).xy * win_size;
				vec2 p3_2 = (screens[12] - screens[13]).xy * win_size;

				vec4 out_factors = vec4(
					length(p0_0) + length(p0_1) + length(p0_2),
					length(p1_0) + length(p1_1) + length(p1_2),
					length(p2_0) + length(p2_1) + length(p2_2),
					length(p3_0) + length(p3_1) + length(p3_2));

				// ..., now the number of splits, clamp it to some reasonable numbers
				out_factors /= triangle_size;
				out_factors = clamp(out_factors, 3.0f, 30.0f);

				// Set the tessellation levels
				gl_TessLevelInner[1] = (out_factors[0] + out_factors[2]) / 2.0;
				gl_TessLevelInner[0] = (out_factors[1] + out_factors[3]) / 2.0;
				gl_TessLevelOuter[0] = out_factors[0];
				gl_TessLevelOuter[1] = out_factors[1];
				gl_TessLevelOuter[2] = out_factors[2];
				gl_TessLevelOuter[3] = out_factors[3];
			}
		}
	}
}