#version 400

layout(quads, fractional_odd_spacing, ccw) in;

layout (std140) uniform Matrices
{
	mat4 m_model;
	mat4 m_view;
	mat4 m_projection;
	mat4 m_pvm;
};

uniform sampler2D bumpTex;
uniform float height_scale;

out Data
{
	vec3 Position;
	vec3 Normal;
	vec3 Tangent;
	vec3 Bitangent;
	vec2 TexCoord;
} DataOut;

in Data
{
	vec4 position;
} DataIn[];

// Evaluates the bezier patch
vec3 bezier(in vec3 v1, in vec3 v2, in vec3 v3, in vec3 v4, in float f)
{
	float of = 1.0 - f;
	return v1*of*of*of + 3.0*v2*f*of*of + 3.0*v3*f*f*of + v4*f*f*f;
}

// Evaluates the derivation of the bezier patch
vec3 bezier_der(in vec3 v1, in vec3 v2, in vec3 v3, in vec3 v4, in float f)
{
	float of = 1.0 - f;
	return -v1*3.0*of*of + 3.0*v2*(of*of-2.0*of*f) + 3.0*v3*(-f*f+2.0*f*of) + 3.0*v4*f*f;
}

//----------------------------------------------------------------
// TODO:
// 6) Compute the position of the tessellated vertex
// 6) Transform the tessellated vertex
// 7) Uncomment computation of the normal of the tessellated vertex
// 8) Uncomment fix of the normal
// 9) Displace the vertex position
// 10) Pass texture coordinates to FS
// 10) Pass tangent and bitangent without normalizing (we need the length) to FS
//----------------------------------------------------------------

void main()
{
	// Only to make it more readable
	vec3 ip[16];
	for (int i = 0; i < 16; i++) ip[i] = DataIn[i].position.xyz;

	// TODO 6) Compute the position of the tessellated vertex
	vec3 v0 = bezier(ip[0], ip[1], ip[2], ip[3], gl_TessCoord.x);
	vec3 v1 = bezier(ip[4], ip[5], ip[6], ip[7], gl_TessCoord.x);
	vec3 v2 = bezier(ip[8], ip[9], ip[10], ip[11], gl_TessCoord.x);
	vec3 v3 = bezier(ip[12], ip[13], ip[14], ip[15], gl_TessCoord.x);
	DataOut.Position = bezier(v0, v1, v2, v3, gl_TessCoord.y);

	// TODO 7) Uncomment computation of the normal of the tessellated vertex
	// Compute the tangent and bitangent
	vec3 v0_dx = bezier_der(ip[0], ip[1], ip[2], ip[3], gl_TessCoord.x);
	vec3 v1_dx = bezier_der(ip[4], ip[5], ip[6], ip[7], gl_TessCoord.x);
	vec3 v2_dx = bezier_der(ip[8], ip[9], ip[10], ip[11], gl_TessCoord.x);
	vec3 v3_dx = bezier_der(ip[12], ip[13], ip[14], ip[15], gl_TessCoord.x);
	vec3 v_dx = bezier(v0_dx, v1_dx, v2_dx, v3_dx, gl_TessCoord.y);
	vec3 v_dy = bezier_der(v0, v1, v2, v3, gl_TessCoord.y);

	// TODO 8) Uncomment fix of the normal
	// Fix the tangent and bitangent if a patch's side is degenerated into a point
	if (dot(v_dx, v_dx) < 1e-3)
	{
		v0 = bezier(ip[0], ip[1], ip[2], ip[3], 1.0);
		v1 = bezier(ip[4], ip[5], ip[6], ip[7], 1.0);
		v2 = bezier(ip[8], ip[9], ip[10], ip[11], 1.0);
		v3 = bezier(ip[12], ip[13], ip[14], ip[15], 1.0);
		v_dx = bezier_der(v0, v1, v2, v3, gl_TessCoord.y);
		v0 = bezier(ip[0], ip[1], ip[2], ip[3], 0.0);
		v1 = bezier(ip[4], ip[5], ip[6], ip[7], 0.0);
		v2 = bezier(ip[8], ip[9], ip[10], ip[11], 0.0);
		v3 = bezier(ip[12], ip[13], ip[14], ip[15], 0.0);
		v_dy = bezier_der(v0, v1, v2, v3, gl_TessCoord.y);
	}
	else if (dot(v_dy, v_dy) < 1e-3)
	{
		v_dx = bezier(v0_dx, v1_dx, v2_dx, v3_dx, 0.0f);
		v_dy = bezier(v0_dx, v1_dx, v2_dx, v3_dx, 1.0f);
	}

	// TODO 7) Uncomment computation of the normal of the tessellated vertex
	DataOut.Normal = normalize(cross(v_dx, v_dy));

	// TODO 10) Pass texture coordinates to FS
	DataOut.TexCoord = gl_TessCoord.xy;
	
	// TODO 10) Pass tangent and bitangent without normalizing to FS
	DataOut.Tangent = v_dx;
	DataOut.Bitangent = v_dy;

	// TODO 9) Displace the vertex position
	float height = textureLod(bumpTex, DataOut.TexCoord, 0.0).x;
	DataOut.Position += DataOut.Normal * height * height_scale;

	// TODO 6) Transform the tessellated vertex (hint: use m_projection)
	gl_Position = m_projection * vec4(DataOut.Position, 1.0);
}